@@ -113,6 +113,7 @@ class Client(AuthenticationClient):
         self.add_event_callback(self.on_member, nio.RoomMemberEvent)  # type:ignore
         self.add_event_callback(self.on_redact, nio.RedactionEvent)  # type:ignore
         self.add_event_callback(self.on_reaction, nio.ReactionEvent)  # type:ignore
+        self.add_event_callback(self.on_power_levels, nio.PowerLevelsEvent)  # type:ignore
         self.add_presence_callback(self.on_presence, nio.PresenceEvent)  # type:ignore
         self.add_ephemeral_callback(self.on_receipt, nio.ReceiptEvent)  # type:ignore
         self.add_ephemeral_callback(
@@ -253,8 +254,10 @@ class Client(AuthenticationClient):
     async def on_member(self, room: nio.MatrixRoom, event: nio.RoomMemberEvent):
         muc = await self.__get_muc(room)
         participant = await self.get_participant(room, event)
-        if event.membership == "join":
-            pass
+        if event.membership == "join" and participant.contact:
+            participant.set_affiliation_from_power_level(
+                room.power_levels.get_user_level(participant.contact.legacy_id)
+            )
         elif event.membership == "leave":
             if participant.is_user:
                 await self.session.bookmarks.remove(muc)
@@ -292,8 +295,12 @@ class Client(AuthenticationClient):
         sender = source["sender"]
         emoji = event.key
 
-        await self.reactions.add(room.room_id, msg_id, sender, emoji, event.event_id)
-        reactions = await self.reactions.get(room.room_id, msg_id, sender)
+        with self.reactions.session() as s:
+            await self.reactions.add(
+                s, room.room_id, msg_id, sender, emoji, event.event_id
+            )
+            reactions = await self.reactions.get(s, room.room_id, msg_id, sender)
+            s.commit()
 
         participant = await self.get_participant(room, event)
         participant.react(msg_id, reactions, **kw)
@@ -302,13 +309,15 @@ class Client(AuthenticationClient):
     async def on_redact(self, room: nio.MatrixRoom, event: nio.RedactionEvent):
         self.log.debug("Redaction: %s", event)
         redacter = await self.get_participant(room, event)
-        if reaction_target := self.reactions.remove(event.redacts):
-            msg_id = await self.get_original_id(room.room_id, reaction_target.event)
-            reactions = await self.reactions.get(
-                reaction_target.room, msg_id, reaction_target.sender
-            )
-            redacter.react(msg_id, reactions)
-            return
+        with self.reactions.session() as s:
+            if reaction_target := self.reactions.remove(s, room.room_id, event.redacts):
+                msg_id = await self.get_original_id(room.room_id, reaction_target.event)
+                reactions = await self.reactions.get(
+                    s, reaction_target.room, msg_id, event.sender
+                )
+                redacter.react(msg_id, reactions)
+                s.commit()
+                return
 
         redacted_event = await self.get_event(room.room_id, event.redacts)
 
@@ -317,6 +326,12 @@ class Client(AuthenticationClient):
         else:
             redacter.moderate(event.redacts, event.reason)
 
+    async def on_power_levels(self, room: nio.MatrixRoom, event: nio.PowerLevelsEvent):
+        muc = await self.__get_muc(room)
+        for user_id, power_level in event.power_levels.users.items():
+            participant = await muc.get_participant_by_legacy_id(user_id)
+            participant.set_affiliation_from_power_level(power_level)
+
     @alru_cache(maxsize=1000)
     async def get_original_id(self, room_id: str, event_id: str) -> str:
         event = await self.get_event(room_id, event_id)
